﻿Imports System.Windows.Forms

Public Class AverDTW

    Private Const DDIAG As Integer = 0                 ' moving direction in the dynamic time warping matrix - diagonal
    Private Const DDOWN As Integer = 1                 ' moving direction in the dynamic time warping matrix - down
    Private Const DRIGHT As Integer = 2                ' moving direction in the dynamic time warping matrix - right

    ' Properties
    Private intComponent As Integer             ' number of components for one curve
    Private lngNumOfIntervals As Integer        ' number of time intervals for one curve
    Private intWinConstr As Integer             ' windows constraint (percent)

    ' original-curves
    Private OrigCurves(,,) As Double            ' array with original curves
    Private anzOrigPoints() As Long             ' point number for (smoothed) original curves
    Private OrigSmooth(,,) As Double            ' array with smoothed original curves
    Private OrigSmEq(,,) As Double              ' array with smoothed equidistant original curves
    Private dblDerivOrig(,,) As Double          ' 1st derivative OrigSmEq()
    Private MaxValue(,) As Double               ' maximal value of each component
    Private MaxDeriv(,) As Double               ' maximal value of the 1st derivative of each component

    Private AverCurve(,) As Double              ' average curve
    Private DerivAverCurve(,) As Double         ' 1st derivative (avearage curve)
    Private AverMaxValue() As Double            ' max values of the components (avearage curve)
    Private AverMaxDeriv() As Double            ' max values of the 1st derivative (avearage curve)

    Private AverCurveSm(,) As Double            ' smoothed average curve
    Private DerivAverCurveSm(,) As Double       ' 1st derivative (smoothed average curve)
    Private AverMaxValueSm() As Double          ' max values (smoothed average curve)
    Private AverMaxDerivSm() As Double          ' max values of 1st derivative (smoothed average curve)

    Private MaxCurve(,) As Double               ' max values curve
    Private MinCurve(,) As Double               ' min values curve
    Private MedianCurve(,) As Double            ' median curve
    Private Perc25Curve(,) As Double            ' 25 percentil curve
    Private Perc75Curve(,) As Double            ' 75 percentil curve

    Private MaxCurveSm(,) As Double             ' smoothed max values curve
    Private MinCurveSm(,) As Double             ' smoothed min values curve
    Private MedianCurveSm(,) As Double          ' smoothed median curve
    Private Perc25CurveSm(,) As Double          ' smoothed 25 percentil curve
    Private Perc75CurveSm(,) As Double          ' smoothed 75 percentil curve

    Private dblSmooth As Double                 ' smoothing factor for average curve (percent)
    Private dblSmoothMedian As Double           ' smoothing factor for median curve (percent)
    Private dblSmooth25Perc As Double           ' smoothing factor for 25 percentil curve (percent)
    Private dblSmooth75Perc As Double           ' smoothing factor for 75 percentil curve (percent)
    Private dblSmoothMax As Double              ' smoothing factor for max values curve (percent)
    Private dblSmoothMin As Double              ' smoothing factor for min values curve (percent)

    Private sComment As String                  ' comment of evaluation
    Private blnReady As Boolean                 ' = true - evaluation is ready
    Private iError As Integer                   ' <> 0 - error

#Region "Properties"

    ''' <summary>
    '''  = true - evaluation is ready
    ''' </summary>
    ''' <returns></returns>
    Public ReadOnly Property CalcReady() As Boolean
        Get
            Return blnReady
        End Get
    End Property

    ''' <summary>
    ''' > Error: = 0 - no error; >0 - error in calculation
    ''' </summary>
    ''' <returns></returns>
    Public ReadOnly Property AverDtwError() As Integer
        Get
            Return iError
        End Get
    End Property

    ''' <summary>
    ''' number of component for dtw path
    ''' </summary>
    ''' <returns></returns>
    Public Property DTWComp() As Integer
        Get
            Return intComponent
        End Get
        Set(ByVal iNewValue As Integer)
            intComponent = iNewValue
        End Set
    End Property

    ''' <summary>
    ''' number of time intervals for one curve
    ''' </summary>
    ''' <returns></returns>
    Public Property NumOfIntervals() As Integer
        Get
            Return lngNumOfIntervals
        End Get
        Set(ByVal lNewValue As Integer)
            If lNewValue <= 0 Then
                MsgBox("Count of intervals has to be a positive number!", MsgBoxStyle.Exclamation, "ERROR")
                Exit Property
            End If
            lngNumOfIntervals = lNewValue
        End Set
    End Property

    ''' <summary>
    ''' windows constraint (percent)
    ''' </summary>
    ''' <returns></returns>
    Public Property WinConstraint() As Integer
        Get
            Return intWinConstr
        End Get
        Set(ByVal vNewValue As Integer)
            If vNewValue < 0 Or vNewValue > 100 Then
                MsgBox("The value of the window constraint can only lie between the numbers 0 and 100!",
                        vbExclamation, "ERROR")
                Exit Property
            End If
            intWinConstr = vNewValue
        End Set
    End Property

    ''' <summary>
    ''' smoothing factor for average curve (percent)
    ''' </summary>
    ''' <returns></returns>
    Public Property SmoothAver() As Double
        Get
            Return dblSmooth
        End Get
        Set(ByVal vNewValue As Double)
            If vNewValue <= 0 Or vNewValue > 100 Then
                MsgBox("The value of the smoothing factor can only lie between the numbers 0 and 100!",
                        vbExclamation, "ERROR")
                Exit Property
            End If
            dblSmooth = vNewValue
        End Set
    End Property

    ''' <summary>
    ''' smoothing factor for median curve (percent)
    ''' </summary>
    ''' <returns></returns>
    Public Property SmoothMedian() As Double
        Get
            Return dblSmoothMedian
        End Get
        Set(ByVal vNewValue As Double)
            If vNewValue <= 0 Or vNewValue > 100 Then
                MsgBox("The value of the smoothing factor can only lie between the numbers 0 and 100!",
                        vbExclamation, "ERROR")
                Exit Property
            End If
            dblSmoothMedian = vNewValue
        End Set
    End Property

    ''' <summary>
    ''' smoothing factor for 25 percentil curve (percent)
    ''' </summary>
    ''' <returns></returns>
    Public Property Smooth25Perc() As Double
        Get
            Return dblSmooth25Perc
        End Get
        Set(ByVal vNewValue As Double)
            If vNewValue <= 0 Or vNewValue > 100 Then
                MsgBox("The value of the smoothing factor can only lie between the numbers 0 and 100!",
                        vbExclamation, "ERROR")
                Exit Property
            End If
            dblSmooth25Perc = vNewValue
        End Set
    End Property

    ''' <summary>
    ''' smoothing factor for 75 percentil curve (percent)
    ''' </summary>
    ''' <returns></returns>
    Public Property Smooth75Perc() As Double
        Get
            Return dblSmooth75Perc
        End Get
        Set(ByVal vNewValue As Double)
            If vNewValue <= 0 Or vNewValue > 100 Then
                MsgBox("The value of the smoothing factor can only lie between the numbers 0 and 100!",
                        vbExclamation, "ERROR")
                Exit Property
            End If
            dblSmooth75Perc = vNewValue
        End Set
    End Property

    ''' <summary>
    ''' smoothing factor for max values curve (percent)
    ''' </summary>
    ''' <returns></returns>
    Public Property SmoothMax() As Double
        Get
            Return dblSmoothMax
        End Get
        Set(ByVal vNewValue As Double)
            If vNewValue <= 0 Or vNewValue > 100 Then
                MsgBox("The value of the smoothing factor can only lie between the numbers 0 and 100!",
                        vbExclamation, "ERROR")
                Exit Property
            End If
            dblSmoothMax = vNewValue
        End Set
    End Property

    ''' <summary>
    ''' smoothing factor for min values curve (percent)
    ''' </summary>
    ''' <returns></returns>
    Public Property SmoothMin() As Double
        Get
            Return dblSmoothMin
        End Get
        Set(ByVal vNewValue As Double)
            If vNewValue <= 0 Or vNewValue > 100 Then
                MsgBox("The value of the smoothing factor can only lie between the numbers 0 and 100!",
                        vbExclamation, "ERROR")
                Exit Property
            End If
            dblSmoothMin = vNewValue
        End Set
    End Property

    ''' <summary>
    ''' evalueted curves: average values (read only)
    ''' </summary>
    ''' <returns></returns>
    Public ReadOnly Property AverDtw() As Double(,)
        Get
            If blnReady Then AverDtw = AverCurveSm Else ReDim AverDtw(0, 0)
        End Get
    End Property

    ''' <summary>
    ''' evalueted curves: min values (read only)
    ''' </summary>
    ''' <returns></returns>
    Public ReadOnly Property MinDtw() As Double(,)
        Get
            If blnReady Then MinDtw = MinCurveSm Else ReDim MinDtw(0, 0)
        End Get
    End Property

    ''' <summary>
    ''' evalueted curves: max values (read only)
    ''' </summary>
    ''' <returns></returns>
    Public ReadOnly Property MaxDtw() As Double(,)
        Get
            If blnReady Then MaxDtw = MaxCurveSm Else ReDim MaxDtw(0, 0)
        End Get
    End Property

    ''' <summary>
    ''' evalueted curves: median values (read only)
    ''' </summary>
    ''' <returns></returns>
    Public ReadOnly Property MedianDtw() As Double(,)
        Get
            If blnReady Then MedianDtw = MedianCurveSm Else ReDim MedianDtw(0, 0)
        End Get
    End Property

    ''' <summary>
    ''' evalueted curves: 25-percentil (read only)
    ''' </summary>
    ''' <returns></returns>
    Public ReadOnly Property Perc25Dtw() As Double(,)
        Get
            If blnReady Then Perc25Dtw = Perc25CurveSm Else ReDim Perc25Dtw(0, 0)
        End Get
    End Property

    ''' <summary>
    ''' evalueted curves: 75-percentil (read only)
    ''' </summary>
    ''' <returns></returns>
    Public ReadOnly Property Perc75Dtw() As Double(,)
        Get
            If blnReady Then Perc75Dtw = Perc75CurveSm Else ReDim Perc75Dtw(0, 0)
        End Get
    End Property

    ''' <summary>
    ''' comment to evaluation progress
    ''' </summary>
    ''' <returns></returns>
    Public ReadOnly Property Comment() As String
        Get
            Return sComment
        End Get
    End Property
#End Region

    ''' <summary>
    ''' constructor
    ''' </summary>
    Public Sub New()

        ' standard values
        intComponent = 1           ' 1st component for dynamic time warping path
        lngNumOfIntervals = 100    ' 100 time intervals
        intWinConstr = 20          ' 20% window constraint

        ' smoothing for the output curves
        dblSmooth = 100.0
        dblSmoothMedian = 100.0
        dblSmooth25Perc = 100.0
        dblSmooth75Perc = 100.0
        dblSmoothMax = 100.0
        dblSmoothMin = 100.0

        ' all arrays
        ReDim OrigCurves(0, 0, 0)
        ReDim OrigSmooth(0, 0, 0)
        ReDim OrigSmEq(0, 0, 0)
        ReDim anzOrigPoints(0)

        ReDim dblDerivOrig(0, 0, 0)
        ReDim MaxValue(0, 0)
        ReDim MaxDeriv(0, 0)

        ReDim AverCurve(0, 0)
        ReDim DerivAverCurve(0, 0)
        ReDim AverMaxValue(0)
        ReDim AverMaxDeriv(0)

        ReDim MaxCurve(0, 0)
        ReDim MinCurve(0, 0)
        ReDim MedianCurve(0, 0)
        ReDim Perc25Curve(0, 0)
        ReDim Perc75Curve(0, 0)

        ReDim AverCurveSm(0, 0)
        ReDim DerivAverCurveSm(0, 0)
        ReDim AverMaxValueSm(0)
        ReDim AverMaxDerivSm(0)

        ReDim MaxCurveSm(0, 0)
        ReDim MinCurveSm(0, 0)
        ReDim MedianCurveSm(0, 0)
        ReDim Perc25CurveSm(0, 0)
        ReDim Perc75CurveSm(0, 0)

        sComment = ""
        blnReady = False
        iError = 0

    End Sub

    ''' <summary>
    ''' destructor
    ''' </summary>
    Protected Overrides Sub Finalize()

        ' all arrays free
        OrigCurves = Nothing
        OrigSmooth = Nothing
        OrigSmEq = Nothing
        anzOrigPoints = Nothing

        dblDerivOrig = Nothing
        MaxValue = Nothing
        MaxDeriv = Nothing

        AverCurve = Nothing
        DerivAverCurve = Nothing
        AverMaxValue = Nothing
        AverMaxDeriv = Nothing

        MaxCurve = Nothing
        MinCurve = Nothing
        MedianCurve = Nothing
        Perc25Curve = Nothing
        Perc75Curve = Nothing

        AverCurveSm = Nothing
        DerivAverCurveSm = Nothing
        AverMaxValueSm = Nothing
        AverMaxDerivSm = Nothing

        MaxCurveSm = Nothing
        MinCurveSm = Nothing
        MedianCurveSm = Nothing
        Perc25CurveSm = Nothing
        Perc75CurveSm = Nothing

        sComment = ""
        iError = 0

        MyBase.Finalize()

    End Sub

    ''' <summary>
    ''' start a new calculation
    ''' </summary>
    Public Sub StartNewCalc()
        ' all arrays
        ReDim OrigCurves(0, 0, 0)
        ReDim OrigSmooth(0, 0, 0)
        ReDim OrigSmEq(0, 0, 0)
        ReDim anzOrigPoints(0)

        ReDim dblDerivOrig(0, 0, 0)
        ReDim MaxValue(0, 0)
        ReDim MaxDeriv(0, 0)

        ReDim AverCurve(0, 0)
        ReDim DerivAverCurve(0, 0)
        ReDim AverMaxValue(0)
        ReDim AverMaxDeriv(0)

        ReDim MaxCurve(0, 0)
        ReDim MinCurve(0, 0)
        ReDim MedianCurve(0, 0)
        ReDim Perc25Curve(0, 0)
        ReDim Perc75Curve(0, 0)

        ReDim AverCurveSm(0, 0)
        ReDim DerivAverCurveSm(0, 0)
        ReDim AverMaxValueSm(0)
        ReDim AverMaxDerivSm(0)

        ReDim MaxCurveSm(0, 0)
        ReDim MinCurveSm(0, 0)
        ReDim MedianCurveSm(0, 0)
        ReDim Perc25CurveSm(0, 0)
        ReDim Perc75CurveSm(0, 0)

        sComment = ""
        blnReady = False
        iError = 0

    End Sub

    ''' <summary>
    ''' Calculation of the average curve dblAverCurve() with dynamic time warping
    ''' </summary>
    ''' <param name="dblAverCurve">the average curve</param>
    Public Sub CalculateDTWAver(ByRef dblAverCurve(,) As Double)

        Try

            iError = 0

            ' control of the dtw path component
            If intComponent > OrigSmEq.GetUpperBound(0) Then
                sComment = "Error in the initializing parameter!"
                iError = 1
                Throw New ApplicationException(sComment)
            End If

            ' comment empty
            sComment = ""
            blnReady = False
            ' count of files in the evaluation
            Dim lngFiles As Integer = OrigSmEq.GetUpperBound(2)

            ' write comment about evaluation progress
            sComment = "Calculation of the 1st derivative and max/min values for the original curves"

            ' 1st derivative and absolute max value for each component
            ' and for the 1st derivative of each components
            Calc_OrigCurves_Derivat()

            'Step 1: Calculation of the reference curve
            sComment = "Step 1: Calculation of the reference curve with " + CStr(lngNumOfIntervals) + " points"
            Calc_RefCurves()

            'Step 2: start the calculation of the average curve
            sComment = "Step 2: Calculation of the average curve"
            Call Calc_Dtw2AverCurve()

            sComment = "Ready"
            blnReady = True
            dblAverCurve = AverCurveSm

        Catch ex As Exception
            Dim sMsg As String = ErrMsg("CalculateDTWAver", "averdtw.dll", ex.Message)
            MsgBox(sMsg, MsgBoxStyle.OkOnly, MsgBoxStyle.Exclamation)
            iError = 1
        End Try
    End Sub

    ''' <summary>
    ''' Add one new original curve:
    ''' </summary>
    ''' <param name="dNewCurve">one 2-dim array</param>
    ''' <param name="dSmooth">smoothing factor (=100% - not smoothig)</param>
    Public Sub AddOneCurve(ByRef dNewCurve(,) As Double, Optional ByVal dSmooth As Double = 100)

        Dim dHelp(,,) As Double        ' help array for curves
        Dim dHelpSmoth(,,) As Double   ' help array for smoothedn curves
        Dim dHelpSmEq(,,) As Double    ' help array for smoothedn equidistant curves

        Dim dblSmooth(,) As Double     ' smoothed curve
        Dim dblSmoothEq(,) As Double   ' smoothed equdistant curve

        Try

            ' intializing
            sComment = ""
            blnReady = False
            iError = 0

            If UBound(dNewCurve, 1) = 0 Or UBound(dNewCurve, 2) = 0 Then
                iError = 1
                Throw New ApplicationException("A 2-dim array is beeing expected!")
            End If

            If UBound(OrigCurves, 1) = 0 Then
                ' 1st curve
                ReDim OrigCurves(dNewCurve.GetUpperBound(0), dNewCurve.GetUpperBound(1), 0)
                ReDim anzOrigPoints(0)
                ReDim OrigSmooth(dNewCurve.GetUpperBound(0), dNewCurve.GetUpperBound(1), 0)
                ReDim OrigSmEq(dNewCurve.GetUpperBound(0), lngNumOfIntervals, 0)
            Else
                ' add a new curve
                ReDim Preserve anzOrigPoints(anzOrigPoints.GetUpperBound(0) + 1)

                If UBound(dNewCurve, 2) <= UBound(OrigCurves, 2) Then
                    ReDim Preserve OrigCurves(dNewCurve.GetUpperBound(0), OrigCurves.GetUpperBound(1), anzOrigPoints.GetUpperBound(0))
                    ReDim Preserve OrigSmooth(dNewCurve.GetUpperBound(0), OrigCurves.GetUpperBound(1), anzOrigPoints.GetUpperBound(0))
                    ReDim Preserve OrigSmEq(dNewCurve.GetUpperBound(0), lngNumOfIntervals, anzOrigPoints.GetUpperBound(0))
                Else
                    ' dimensioning of the help arrays
                    ReDim dHelp(OrigCurves.GetUpperBound(0), OrigCurves.GetUpperBound(1), OrigCurves.GetUpperBound(2))
                    ReDim dHelpSmoth(OrigSmooth.GetUpperBound(0), OrigSmooth.GetUpperBound(1), OrigSmooth.GetUpperBound(2))
                    ReDim dHelpSmEq(OrigSmEq.GetUpperBound(0), OrigSmEq.GetUpperBound(1), OrigSmEq.GetUpperBound(2))

                    ' saving the original data and original smooothed data in the help arrays
                    For m As Integer = 0 To OrigCurves.GetUpperBound(2)
                        For i As Integer = 0 To OrigCurves.GetUpperBound(0)
                            For j As Integer = 0 To OrigCurves.GetUpperBound(1)
                                dHelp(i, j, m) = OrigCurves(i, j, m)
                                dHelpSmoth(i, j, m) = OrigSmooth(i, j, m)
                            Next
                            For j As Integer = 0 To lngNumOfIntervals
                                dHelpSmEq(i, j, m) = OrigSmEq(i, j, m)
                            Next
                        Next
                    Next

                    ' redimensioning of the arrays with the original data and original smooothed data
                    ReDim OrigCurves(OrigCurves.GetUpperBound(0), dNewCurve.GetUpperBound(1), anzOrigPoints.GetUpperBound(0))
                    ReDim OrigSmooth(OrigSmooth.GetUpperBound(0), dNewCurve.GetUpperBound(1), anzOrigPoints.GetUpperBound(0))
                    ReDim OrigSmEq(OrigSmEq.GetUpperBound(0), lngNumOfIntervals, anzOrigPoints.GetUpperBound(0))

                    ' saving back the original data and original smooothed data
                    For m As Integer = 0 To OrigCurves.GetUpperBound(2) - 1
                        For i As Integer = 0 To OrigCurves.GetUpperBound(0)
                            For j As Integer = 0 To CInt(anzOrigPoints(m))
                                OrigCurves(i, j, m) = dHelp(i, j, m)
                                OrigSmooth(i, j, m) = dHelpSmoth(i, j, m)
                            Next

                            For j As Integer = 0 To lngNumOfIntervals
                                OrigSmEq(i, j, m) = dHelpSmEq(i, j, m)
                            Next
                        Next
                    Next

                    ' memory free
                    ReDim dHelp(0, 0, 0)
                    ReDim dHelpSmoth(0, 0, 0)
                    ReDim dHelpSmEq(0, 0, 0)
                End If
            End If

            ' add new curve
            Dim k As Integer = OrigCurves.GetUpperBound(2)
            anzOrigPoints(k) = dNewCurve.GetUpperBound(1)
            For i = 0 To dNewCurve.GetUpperBound(0)
                For j = 0 To dNewCurve.GetUpperBound(1)
                    OrigCurves(i, j, k) = dNewCurve(i, j)
                Next j
            Next i

            ' smoothing
            ReDim dblSmooth(0, 0)
            Calc_CurveSmoothing(dNewCurve, dblSmooth, dSmooth)

            For i As Integer = 0 To dblSmooth.GetUpperBound(0)
                For j As Integer = 0 To dblSmooth.GetUpperBound(1)
                    OrigSmooth(i, j, k) = dblSmooth(i, j)
                Next
            Next

            ' equidistant segmentation of the original data
            ReDim dblSmoothEq(0, 0)
            Calc_CurveEquidist(dblSmooth, dblSmoothEq)

            For i As Integer = 0 To dblSmoothEq.GetUpperBound(0)
                For j As Integer = 0 To dblSmoothEq.GetUpperBound(1)
                    OrigSmEq(i, j, k) = dblSmoothEq(i, j)
                Next
            Next

        Catch ex As Exception
            Dim sMsg As String = ErrMsg("AddOneCurve", "averdtw.dll", ex.Message)
            MsgBox(sMsg, MsgBoxStyle.OkOnly, MsgBoxStyle.Exclamation)
            iError = 1
        End Try
    End Sub

    ''' <summary>
    ''' Smothing one curve
    ''' </summary>
    ''' <param name="dblArray">2dim array</param>
    ''' <param name="dblarraySmooth">smoothed 2-dim array </param>
    ''' <param name="dblSmoothing">smoothing factor in percent</param>
    Private Sub Calc_CurveSmoothing(ByRef dblArray(,) As Double, ByRef dblarraySmooth(,) As Double,
                                        ByVal dblSmoothing As Double)

        Try

            If UBound(dblArray) = 0 Then Throw New ApplicationException("Array for smoothing is empty! ")

            'initialising
            Dim TeilNeu As Double = dblSmoothing / 100.0#          ' value between 0.01 ... 0.99
            Dim TeilAlt As Double = 1 - TeilNeu                    ' value between 0.01 ... 0.99

            ' size of the 1st dimension of dblarray()
            Dim lngDim1 As Integer = dblArray.GetUpperBound(0)
            ' size of the 2d dimension of dblarray()
            Dim lngDim2 As Integer = dblArray.GetUpperBound(1)

            ReDim dblarraySmooth(lngDim1, lngDim2)

            ' preserve the time
            dblarraySmooth(0, 0) = dblArray(0, 0)
            ' save the 1st value of each component
            For i As Integer = 1 To dblArray.GetUpperBound(0)
                dblarraySmooth(i, 0) = dblArray(i, 0)
            Next

            'forward smoothing
            For j As Integer = 1 To dblArray.GetUpperBound(1)
                ' preserve the time
                dblarraySmooth(0, j) = dblArray(0, j)
                ' forward smoothing of the components
                For i As Integer = 1 To dblArray.GetUpperBound(0)
                    dblarraySmooth(i, j) = dblarraySmooth(i, j - 1) * TeilAlt + dblArray(i, j) * TeilNeu
                Next
            Next

            'backword smoothing
            For j As Integer = dblArray.GetUpperBound(1) - 1 To 0 Step -1
                ' backward smoothing of the components
                For i As Integer = 1 To dblArray.GetUpperBound(0)
                    dblarraySmooth(i, j) = dblarraySmooth(i, j + 1) * TeilAlt + dblarraySmooth(i, j) * TeilNeu
                Next
            Next

        Catch ex As Exception
            iError = 1
            Dim sMsg As String = ErrMsg("Calc_CurveSmoothing", "averdtw.dll", "Error while smoothing the data!" & ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Sub

    ''' <summary>
    ''' Resampling the original data for constant time steps
    ''' </summary>
    ''' <param name="dblArray">array with original data</param>
    ''' <param name="dblArrEquidist">array with time-equidistant data</param>
    Private Sub Calc_CurveEquidist(ByRef dblArray(,) As Double, ByRef dblArrEquidist(,) As Double)

        Try
            Dim dblComponent() As Double    ' a 1-dim array of all components at one point of time
            ' count of recordsets in the array with smoothed data
            Dim lngNumOfSmooth As Integer = UBound(dblArray, 2)

            ' max value of time
            Dim dblTimeMax As Double = dblArray(0, lngNumOfSmooth)

            ' evaluate the time step
            Dim dblDeltaT As Double = dblTimeMax / lngNumOfIntervals

            ' count of components in one recordset
            Dim intNumOfComp As Integer = dblArray.GetUpperBound(0)

            ' redimensioning of array
            ReDim dblArrEquidist(intNumOfComp, lngNumOfIntervals)

            ' the 1st point of time = 0 millisekanden
            dblArrEquidist(0, 0) = 0

            For i As Integer = 1 To lngNumOfIntervals
                dblArrEquidist(0, i) = dblArrEquidist(0, 0) + i * dblDeltaT
            Next

            ' search or interpolate all components on the smoothed data
            For i As Integer = 0 To lngNumOfIntervals
                ' one component is the time
                ReDim dblComponent(intNumOfComp)
                Search_ForTime(dblArray, dblArrEquidist(0, i), dblComponent)

                ' save all components
                For j As Integer = 1 To intNumOfComp
                    dblArrEquidist(j, i) = dblComponent(j)
                Next
            Next

        Catch ex As Exception
            iError = 1
            Dim sMsg As String = ErrMsg("Calc_CurveEquidist", "averdtw.dll",
                                        "Error in the equidistant segmentation of the data! " & ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Sub

    ''' <summary>
    '''  Search for one recordset from the array dblArr() with data for the point of time dblTime
    ''' </summary>
    ''' <param name="dblArr"> time sorted data array</param>
    ''' <param name="dblTime">point of time</param>
    ''' <param name="dblSearch"> one recordset from the array dblArr() with data for the point of time dblTime</param>
    Private Sub Search_ForTime(ByRef dblArr(,) As Double, ByRef dblTime As Double, ByRef dblSearch() As Double)

        Try

            Dim lngNumOfDS As Integer = dblArr.GetUpperBound(1)        ' count of recordsets in array
            Dim intNumOfComp As Integer = dblSearch.GetUpperBound(0)   ' count of components in array

            ' the 1st value in dblArr() is the dblTime
            If dblTime = dblArr(0, 0) Then
                For j As Integer = 1 To intNumOfComp
                    dblSearch(j) = dblArr(j, 0)
                Next
                Exit Sub  ' value is found !
            End If

            ' start at the 1st recordset
            Dim lngStart As Integer = 1
            Dim i As Integer = 0
            Dim dblOben As Double                 ' upper boand
            Dim dblUnten As Double                ' lower boand
            ' 1st search bound (step = 50)
            If lngNumOfDS > 51 Then
                For i = 51 To lngNumOfDS Step 50    '
                    dblOben = dblArr(0, i)
                    If dblOben >= dblTime Then
                        ' change the start number
                        lngStart = i - 50
                        Exit For
                    End If
                Next

                If lngStart = 1 And dblArr(0, i - 50) < dblTime Then
                    lngStart = i - 50
                End If
            End If

            ' 2nd search bound (step = 1)
            For i = lngStart To lngNumOfDS
                dblOben = dblArr(0, i)

                Select Case dblOben
                    Case Is < dblTime
                        If i = lngNumOfDS And Math.Abs(dblOben - dblTime) < 0.0001 Then
                            For j As Integer = 1 To intNumOfComp
                                dblSearch(j) = dblArr(j, i)
                            Next
                        End If

                    Case Is = dblTime
                        For j As Integer = 1 To intNumOfComp
                            dblSearch(j) = dblArr(j, i)
                        Next
                        Exit Sub  ' value is found !

                    Case Is >= dblTime
                        ' lower boand
                        dblUnten = dblArr(0, i - 1)
                        For j As Integer = 1 To intNumOfComp
                            ' components for this point of time
                            Dim dblSearchOben As Double = dblArr(j, i)
                            Dim dblSearchUnten As Double = dblArr(j, i - 1)
                            ' interpolate the values for the components
                            dblSearch(j) = dblSearchUnten + (dblTime - dblUnten) * (dblSearchOben - dblSearchUnten) / (dblOben - dblUnten)
                        Next

                        Exit Sub  ' value is found !

                    Case Else
                End Select
            Next

        Catch ex As Exception
            iError = 1
            Dim sMsg As String = ErrMsg("Search_ForTime", "averdtw.dll", ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Sub

    ''' <summary>
    ''' Evaluation of the 1st derivative of all smoothed original curves
    ''' Search the max value for each component and its 1st derivative
    ''' </summary>
    Private Sub Calc_OrigCurves_Derivat()

        Dim dblOrigSmEq(,) As Double      ' array for smoothed equidistant original data
        Dim dblDeriv(,) As Double         ' array for 1st derivative
        Dim dblValueMax() As Double       ' a 1-dim array of max values for all components in one file
        Dim dblDerivMax() As Double       ' a 1-dim array of max value for the 1st derivative of all components in one file

        Try

            ReDim dblDerivOrig(OrigSmEq.GetUpperBound(0), lngNumOfIntervals, OrigSmEq.GetUpperBound(2))
            ReDim MaxValue(OrigSmEq.GetUpperBound(0), OrigSmEq.GetUpperBound(2))
            ReDim MaxDeriv(OrigSmEq.GetUpperBound(0), OrigSmEq.GetUpperBound(2))

            ' equidistant data of all files
            For n As Integer = 0 To OrigSmEq.GetUpperBound(2)
                ' redimension the help array
                ReDim dblOrigSmEq(OrigSmEq.GetUpperBound(0), lngNumOfIntervals)
                ' save the values in the help array
                For j As Integer = 0 To OrigSmEq.GetUpperBound(0)
                    For k As Integer = 0 To lngNumOfIntervals
                        dblOrigSmEq(j, k) = OrigSmEq(j, k, n)
                    Next
                Next

                ' evaluation of the 1st derivative for all measurement points
                ReDim dblDeriv(0, 0)
                Calc_DerivNum(dblOrigSmEq, dblDeriv)

                ' search for max values of each component and its 1st derivative
                ReDim dblValueMax(0)
                Search_Max(dblOrigSmEq, dblValueMax)

                ReDim dblDerivMax(0)
                Search_Max(dblDeriv, dblDerivMax)

                ' i - file number;  j - component number; k - measurement number
                For j As Integer = 0 To dblOrigSmEq.GetUpperBound(0)
                    ' equidistant smoothed original data:
                    ' dblOrigSmEq() - data from one file
                    ' OrigSmEq() - data from all files
                    For k As Integer = 0 To dblOrigSmEq.GetUpperBound(1)
                        dblDerivOrig(j, k, n) = dblDeriv(j, k)
                    Next

                    ' max values of all components
                    If j > 0 Then
                        MaxValue(j, n) = dblValueMax(j)        'j = 0 - time
                        MaxDeriv(j, n) = dblDerivMax(j)
                    End If
                Next
            Next

        Catch ex As Exception
            iError = 1
            Dim sMsg As String = ErrMsg("Calc_OrigCurves_Derivat", "averdtw.dll", ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Sub

    ''' <summary>
    ''' Evaluation of the 1st derivative for an array with smoothedn and equidistant original data
    ''' </summary>
    ''' <param name="dblOrigCurve">array with components form one measurement</param>
    ''' <param name="dblDeriv">array of the 1st derivative for each point of time</param>
    Private Sub Calc_DerivNum(ByRef dblOrigCurve(,) As Double, ByRef dblDeriv(,) As Double)

        Try

            ' count of components
            Dim intComps As Integer = dblOrigCurve.GetUpperBound(0)
            ' count of measurements
            Dim lngMess As Integer = dblOrigCurve.GetUpperBound(1)

            ' calculation of the 1st derivative
            ReDim dblDeriv(intComps, lngMess)
            ' save the time (without changing)
            For j As Integer = 0 To lngMess
                dblDeriv(0, j) = dblOrigCurve(0, j)
            Next

            For i As Integer = 1 To intComps
                For j As Integer = 1 To lngMess - 1
                    'evaluation of the 1st derivative from
                    ' previous and next measurement point
                    dblDeriv(i, j) = ((dblOrigCurve(i, j) - dblOrigCurve(i, j - 1)) +
                                            ((dblOrigCurve(i, j + 1) - dblOrigCurve(i, j - 1)) / 2)) / 2
                Next
                ' for the 1st und the last point
                dblDeriv(i, 0) = dblDeriv(i, 1)
                dblDeriv(i, lngMess) = dblDeriv(i, lngMess - 1)
            Next

        Catch ex As Exception
            iError = 1
            Dim sMsg As String = ErrMsg("Calc_DerivNum", "averdtw.dll", "Error in the evaluation of the 1st derivative! " & ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Sub

    ''' <summary>
    ''' Search for max values for each component
    ''' </summary>
    ''' <param name="dblArray">a 2dim array (0 To intComponents, 0 To intMeasurements)</param>
    ''' <param name="dblMaxComp">a 1-dim array with max values for each component (1st dim)</param>
    Private Sub Search_Max(ByRef dblArray(,) As Double, ByRef dblMaxComp() As Double)

        Try
            ' count of components
            Dim intComps As Integer = dblArray.GetUpperBound(0)
            ' count of meassurements
            Dim lngMess As Integer = dblArray.GetUpperBound(1)

            ' redimension of max values array
            ReDim dblMaxComp(intComps)

            ' search the max values for all components
            For k As Integer = 0 To intComps
                dblMaxComp(k) = Math.Abs(dblArray(k, 0))
                For i As Integer = 1 To lngMess
                    If Math.Abs(dblArray(k, i)) > dblMaxComp(k) Then dblMaxComp(k) = Math.Abs(dblArray(k, i))
                Next
            Next

        Catch ex As Exception
            iError = 1
            Dim sMsg As String = ErrMsg("Search_Max", "averdtw.dll", "Error while seariching for max values! " & ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Sub

    ''' <summary>
    ''' Evaluation of the reference curve
    ''' </summary>
    Private Sub Calc_RefCurves()

        Try

            Dim dblFunc1(,) As Double           ' 2-dim array with time and values of one component(curve 1)
            Dim dblFunc2(,) As Double           ' 2-dim array with time and values of one component(curve 2)
            Dim dF1Deriv(,) As Double           ' 2-dim array with time and values of 1st derivative for one component(curve 1)
            Dim dF2Deriv(,) As Double           ' 2-dim array with time and values of 1st derivative for one componen(curve 2)
            Dim dblMax1 As Double               ' max value of one component on the curve 1
            Dim dblMax2 As Double               ' max value of one component on the curve 2
            Dim dblMaxDeriv1 As Double          ' max value of 1st derivative for one component on the curve 1
            Dim dblMaxDeriv2 As Double          ' max value of 1st derivative for one component on the curve 2
            Dim dblCurve1(,) As Double          ' 2-dim array with time and values of all component for curve 1
            Dim dblCurve2(,) As Double          ' 2-dim array with time and values of all component for curve 2
            Dim dblAlpha As Double              ' parameter

            ' count of all files
            Dim lngFiles As Long = OrigSmEq.GetUpperBound(2)

            ' evaluate a dynamic time warping path for each 2 curves and calculate the
            ' average curve for all curves
            For lngFile1 As Long = 0 To lngFiles - 1            ' 1st curve for the averaging
                Dim lngFile2 As Long = lngFile1 + 1                 ' 2nd curve for the averaging

                ' preparation for averaging
                If lngFile1 = 0 Then
                    ' get the data from one component (curve 1)
                    ReDim dblFunc1(0, 0)
                    ReDim dF1Deriv(0, 0)
                    Get3Dim_1CompForDTW(OrigSmEq, dblDerivOrig, MaxValue, MaxDeriv,
                                             dblFunc1, dF1Deriv, dblMax1, dblMaxDeriv1,
                                             CInt(lngFile1), intComponent)

                    ' get the data from all components (curve 1)
                    ReDim dblCurve1(0, 0)
                    Get3Dim_AllComp(OrigSmEq, dblCurve1, CInt(lngFile1))

                Else
                    ' get the data from one component (curve 2=temporary average curve)
                    ReDim dblFunc1(0, 0)
                    ReDim dF1Deriv(0, 0)
                    Get2Dim_1CompForDTW(AverCurve, DerivAverCurve, AverMaxValue, AverMaxDeriv,
                                        dblFunc1, dF1Deriv, dblMax1, dblMaxDeriv1, intComponent)

                    ' get the data from all components (curve 2=temporary average curve)
                    ReDim dblCurve1(0, 0)
                    Copy2Dim_To_2Dim(AverCurve, dblCurve1)
                End If

                ' get the data from one component (curve 2)
                ReDim dblFunc2(0, 0)
                ReDim dF2Deriv(0, 0)
                Get3Dim_1CompForDTW(OrigSmEq, dblDerivOrig, MaxValue, MaxDeriv,
                                dblFunc2, dF2Deriv, dblMax2, dblMaxDeriv2, CInt(lngFile2), intComponent)

                ' get the data from all components (curve 2)
                ReDim dblCurve2(0, 0)
                Get3Dim_AllComp(OrigSmEq, dblCurve2, CInt(lngFile2))
                Calc_Alfa(dblAlpha, dblFunc1, dblFunc2, dblMax1, dblMax2, dF1Deriv, dF2Deriv, dblMaxDeriv1,
                                     dblMaxDeriv2)

                ReDim AverCurve(0 To UBound(AverCurve, 1), 0 To lngNumOfIntervals)
                ReDim DerivAverCurve(0 To UBound(AverCurve, 1), 0 To lngNumOfIntervals)

                ' calculate a average curve of 2 curves ausrechnen
                Calc_AverTwoCurves(dblAlpha, dblFunc1, dblFunc2, dblMax1, dblMax2, dF1Deriv, dF2Deriv,
                        dblMaxDeriv1, dblMaxDeriv2, dblCurve1, dblCurve2, lngFile1, lngFile2, AverCurve)

                Calc_DerivNum(AverCurve, DerivAverCurve)
                ' search for the max values of each component and its 1st derivative
                Search_Max(AverCurve, AverMaxValue)
                Search_Max(DerivAverCurve, AverMaxDeriv)

                Application.DoEvents()

            Next lngFile1

            ' equidistant segmenting of the referece curve
            Calc_AverCurveEq(AverCurve, CLng(lngNumOfIntervals))

        Catch ex As Exception
            iError = 1
            Dim sMsg As String = ErrMsg("Calc_RefCurves", "averdtw.dll",
                                         "Error in the calculation of the refence curve! " & ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Sub

    ''' <summary>
    ''' Copy the first 2 dimension of a 3-dim array in a 2-dim array (one components)
    ''' + 1st derivative and max/min value
    ''' </summary>
    ''' <param name="dblCurve"></param>
    ''' <param name="dblDerivCurve"></param>
    ''' <param name="dblMaxValue"></param>
    ''' <param name="dblMaxDeriv"></param>
    ''' <param name="dblFunc"></param>
    ''' <param name="dFDeriv"></param>
    ''' <param name="dblMax"></param>
    ''' <param name="dblMaxDer"></param>
    ''' <param name="intCurveNr"></param>
    ''' <param name="intComponent"></param>
    Private Sub Get3Dim_1CompForDTW(ByRef dblCurve(,,) As Double, ByRef dblDerivCurve(,,) As Double,
                                    ByRef dblMaxValue(,) As Double, ByRef dblMaxDeriv(,) As Double,
                                    ByRef dblFunc(,) As Double, ByRef dFDeriv(,) As Double, ByRef dblMax As Double,
                                    ByRef dblMaxDer As Double, ByRef intCurveNr As Integer, ByRef intComponent As Integer)

        Try

            ReDim dblFunc(0 To 1, 0 To UBound(dblCurve, 2))
            ReDim dFDeriv(0 To 1, 0 To UBound(dblCurve, 2))

            For i As Integer = 0 To UBound(dblCurve, 2)
                dblFunc(0, i) = dblCurve(0, i, intCurveNr)                     ' point of time
                dblFunc(1, i) = dblCurve(intComponent, i, intCurveNr)          ' component for the dtw path evaluation
                dFDeriv(0, i) = dblDerivCurve(0, i, intCurveNr)                ' point of time for 1st derivative array
                dFDeriv(1, i) = dblDerivCurve(intComponent, i, intCurveNr)     ' 1st derivative of the component
            Next

            dblMax = dblMaxValue(intComponent, intCurveNr)                      ' max value for one component
            dblMaxDer = dblMaxDeriv(intComponent, intCurveNr)                   ' max value for the 1st derivative

        Catch ex As Exception
            iError = 1
            Dim sMsg As String = ErrMsg("Get3Dim_1CompForDTW", "averdtw.dll", ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Sub

    ''' <summary>
    ''' Copy the first 2 dimensions of a 3-dim array into a 2-dim array (all components)
    ''' </summary>
    ''' <param name="dblCurve3Dim"></param>
    ''' <param name="dblCurve2Dim"></param>
    ''' <param name="intCurveNr"></param>
    Private Sub Get3Dim_AllComp(ByRef dblCurve3Dim(,,) As Double, ByRef dblCurve2Dim(,) As Double,
                                    ByRef intCurveNr As Integer)

        Try
            ' array redimension
            ReDim dblCurve2Dim(0 To UBound(dblCurve3Dim, 1), 0 To UBound(dblCurve3Dim, 2))

            ' save the values
            For i As Integer = 0 To UBound(dblCurve3Dim, 2)
                For j As Integer = 0 To UBound(dblCurve3Dim, 1)
                    dblCurve2Dim(j, i) = dblCurve3Dim(j, i, intCurveNr)
                Next
            Next

        Catch ex As Exception
            iError = 1
            Dim sMsg As String = ErrMsg("Get3Dim_AllComp", "averdtw.dll", ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Sub

    ''' <summary>
    ''' Copy a 2-dim array into a 2-dim array (one components) + 1st derivative and max/min value
    ''' </summary>
    ''' <param name="dblCurve"></param>
    ''' <param name="dblDerivCurve"></param>
    ''' <param name="dblCurveMax"></param>
    ''' <param name="dblCurDerivMax"></param>
    ''' <param name="dblFunc"></param>
    ''' <param name="dFDeriv"></param>
    ''' <param name="dblMax"></param>
    ''' <param name="dblMaxDeriv"></param>
    ''' <param name="intComponent"></param>
    Private Sub Get2Dim_1CompForDTW(ByRef dblCurve(,) As Double, ByRef dblDerivCurve(,) As Double,
                                  ByRef dblCurveMax() As Double, ByRef dblCurDerivMax() As Double,
                                  ByRef dblFunc(,) As Double, ByRef dFDeriv(,) As Double, ByRef dblMax As Double,
                                  ByRef dblMaxDeriv As Double, ByRef intComponent As Integer)

        Try

            ReDim dblFunc(0 To 1, 0 To UBound(dblCurve, 2))
            ReDim dFDeriv(0 To 1, 0 To UBound(dblDerivCurve, 2))

            For i As Integer = 0 To UBound(dblCurve, 2)
                dblFunc(0, i) = dblCurve(0, i)                  ' point of time
                dblFunc(1, i) = dblCurve(intComponent, i)       ' component for the dtw path evaluation
                dFDeriv(0, i) = dblDerivCurve(0, i)             ' point of time for the 1st derivative array
                dFDeriv(1, i) = dblDerivCurve(intComponent, i)  ' 1st derivative of the component
            Next

            dblMax = dblCurveMax(intComponent)                  ' max value for one component
            dblMaxDeriv = dblCurDerivMax(intComponent)          ' max value for the 1st derivative

        Catch ex As Exception
            iError = 1
            Dim sMsg As String = ErrMsg("Get2Dim_1CompForDTW", "averdtw.dll",
                                        "Error at saving of smoothed (temporary) averaged data! " & ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Sub

    ''' <summary>
    ''' Evaluate an optimal value of the parameter 'alpha'
    ''' </summary>
    ''' <param name="dblAlpha"></param>
    ''' <param name="dblFunc1"></param>
    ''' <param name="dblFunc2"></param>
    ''' <param name="dMaxF1"></param>
    ''' <param name="dMaxF2"></param>
    ''' <param name="dF1Deriv"></param>
    ''' <param name="dF2Deriv"></param>
    ''' <param name="dMaxF1Deriv"></param>
    ''' <param name="dMaxF2Deriv"></param>
    Private Sub Calc_Alfa(ByRef dblAlpha As Double, ByRef dblFunc1(,) As Double, ByRef dblFunc2(,) As Double,
                           ByRef dMaxF1 As Double, ByRef dMaxF2 As Double, ByRef dF1Deriv(,) As Double, ByRef dF2Deriv(,) As Double,
                           ByRef dMaxF1Deriv As Double, ByRef dMaxF2Deriv As Double)

        Try

            Dim lngPath1() As Long              ' dtw-path for the 1st curve
            Dim lngPath2() As Long              ' dtw-path for the 2nd curve
            Dim dblCost(2) As Double       ' the cost function for alpha = 0.3, 0.5 and 0.7
            Dim dblCostMiddl(2) As Double  ' the middle cost function for alpha = 0.3, 0.5 and 0.7

            ' evaluate the cost function for diverst value of alpha
            For k As Integer = 0 To 2
                dblAlpha = (2 * (k + 1) + 1) * 0.1

                ' evaluate the dtw-path
                ReDim lngPath1(0)
                ReDim lngPath2(0)
                Dim lngPathLen As Long = 0
                dblCost(k) = Calc_ComputePath(dblAlpha, dblFunc1, dblFunc2, dMaxF1, dMaxF2, dF1Deriv, dF2Deriv,
                                 dMaxF1Deriv, dMaxF2Deriv, lngPath1, lngPath2, lngPathLen)

                dblCostMiddl(k) = dblCost(k) / lngPathLen
            Next

            ' search for the optimal value of alpha
            Dim dblSumMin As Double = dblCost(0)
            Dim intOptAlfa As Integer = 3        ' optimal value of alpha
            For j As Integer = 1 To 2
                If dblSumMin > dblCost(j) Then
                    dblSumMin = dblCost(j)
                    intOptAlfa = 2 * (j + 1) + 1
                End If
            Next

            dblAlpha = Math.Round(intOptAlfa / 10.0#, 1)

        Catch ex As Exception
            dblAlpha = 0.5
            sComment = "Error while evaluating the 'weighting parameters'(alpha)!"
            iError = 1
            Dim sMsg As String = ErrMsg("Calc_Alfa", "averdtw.dll",
                                        sComment & vbTab & ex.Message)
            Throw New Exception(sMsg, ex)
        End Try

    End Sub

    ''' <summary>
    ''' Evaluation of a Dynamic Time Warping (DTW) path
    ''' Input:    dblFunc1() and dblFunc2() - both curves for DTW-path evaluation
    '''           dMaxF1 and dMaxF2 - max value for one component of the curves 1 and 2
    ''' Output:   DTW-paths for file1 and file2 with measurement numbern and 'cost function'
    '''           lngPath_1() - output-array with dtw path for function 1
    '''           lngPath_2() - output-array with dtw path for function 2
    ''' </summary>
    ''' <param name="dblAlpha"></param>
    ''' <param name="dblFunc1">curve 1 for DTW-path evaluation</param>
    ''' <param name="dblFunc2">curve 2 for DTW-path evaluation</param>
    ''' <param name="dMaxF1">max value for one component of the curves 1</param>
    ''' <param name="dMaxF2">max value for one component of the curves 2</param>
    ''' <param name="dF1Deriv"></param>
    ''' <param name="dF2Deriv"></param>
    ''' <param name="dMaxF1Deriv"></param>
    ''' <param name="dMaxF2Deriv"></param>
    ''' <param name="lngPath_1">output-array with dtw path for function 1</param>
    ''' <param name="lngPath_2">output-array with dtw path for function 2</param>
    ''' <param name="PathLength"></param>
    ''' <returns></returns>
    Private Function Calc_ComputePath(ByVal dblAlpha As Double, ByRef dblFunc1(,) As Double,
            ByRef dblFunc2(,) As Double, ByRef dMaxF1 As Double, ByRef dMaxF2 As Double, ByRef dF1Deriv(,) As Double,
            ByRef dF2Deriv(,) As Double, ByRef dMaxF1Deriv As Double, ByRef dMaxF2Deriv As Double,
            ByRef lngPath_1() As Long, ByRef lngPath_2() As Long, ByRef PathLength As Long) As Double

        Try
            ' window constraint
            Dim intConstr As Integer = CInt((intWinConstr / 100) * lngNumOfIntervals)

            ' calculate the dimension of the 1st and 2nd array
            ' 1st dim dblFunc1()-array (count of recordsets)
            Dim iDimF1 As Integer = dblFunc1.GetUpperBound(1)
            ' 1st dim dblFunc2()-array (count of recordsets)
            Dim iDimF2 As Integer = dblFunc2.GetUpperBound(1)

            ' help array for the evaluation of the paths
            Dim dblMatrix(iDimF1, iDimF2) As Double      ' dtw matrix
            Dim lngPath(iDimF1, iDimF2) As Long               ' help array for the dtw path
            Dim intPathLength(iDimF1, iDimF2) As Integer        ' help array: lenght of the path
            Dim intDownSteps(iDimF1, iDimF2) As Integer      ' help array: count of down steps
            Dim intDiagSteps(iDimF1, iDimF2) As Integer         ' help array: count of diagonal steps
            Dim intRightSteps(iDimF1, iDimF2) As Integer         ' help array: count of right steps

            ' cost function for all three stepe: down, right and diagonal
            Dim dblStpTypeCost(0 To 2) As Double        ' help array: cost function for stepe down, diagonal and right
            For i As Integer = 0 To 2 : dblStpTypeCost(i) = 0 : Next

            ' create starting initialization (no cost free starting gap)
            ' Cost function = Wang&Gasser
            dblMatrix(0, 0) = Calc_CostWG(dblAlpha, dMaxF1, dMaxF2, dblFunc1(1, 0), dblFunc2(1, 0),
                                             dMaxF1Deriv, dMaxF2Deriv, dF1Deriv(1, 0), dF2Deriv(1, 0))

            ' evaluationen of den step DOWN
            For i As Integer = 1 To iDimF1 - 1
                If i < 0.01 * lngNumOfIntervals + 1 Then
                    ' cost function = Wang&Gasser
                    dblMatrix(i, 0) = dblMatrix(i - 1, 0) + Calc_CostWG(dblAlpha, dMaxF1, dMaxF2, dblFunc1(1, i),
                                    dblFunc2(1, 0), dMaxF1Deriv, dMaxF2Deriv, dF1Deriv(1, i), dF2Deriv(1, 0))
                Else
                    ' outside of window constraint
                    dblMatrix(i, 0) = 10 ^ 300     ' infinity
                End If

                intDownSteps(i, 0) = intDownSteps(i - 1, 0) + 1
                lngPath(i, 0) = DDOWN
                intPathLength(i, 0) = intPathLength(i - 1, 0) + 1
            Next

            ' evaluationen of the step RIGHT
            For j As Integer = 1 To iDimF2 - 1
                If j < 0.01 * lngNumOfIntervals + 1 Then
                    dblMatrix(0, j) = dblMatrix(0, j - 1) + Calc_CostWG(dblAlpha, dMaxF1, dMaxF2, dblFunc1(1, 0),
                                    dblFunc2(1, j), dMaxF1Deriv, dMaxF2Deriv, dF1Deriv(1, 0), dF2Deriv(1, j))
                Else
                    dblMatrix(0, j) = 10 ^ 300   ' = infty
                End If

                intRightSteps(0, j) = intRightSteps(0, j - 1) + 1
                lngPath(0, j) = DRIGHT
                intPathLength(0, j) = intPathLength(0, j - 1) + 1
            Next

            ' fill warping dMatrix, from the upper left corner to the right bottom
            For i As Integer = 1 To iDimF1 - 1
                For j As Integer = 1 To iDimF2 - 1
                    If dblMatrix(i - 1, j - 1) = 10 ^ 300 And dblMatrix(i - 1, j) = 10 ^ 300 And dblMatrix(i, j - 1) = 10 ^ 300 Then
                        dblMatrix(i, j) = 10 ^ 300
                    Else

                        ' step_cost = cost_func(i, j, dblFunc1, dblFunc2)
                        ' evaluations of the Step DIAG
                        Dim dblStepCost As Double = Calc_CostWG(dblAlpha, dMaxF1, dMaxF2, dblFunc1(1, i), dblFunc2(1, j),
                                    dMaxF1Deriv, dMaxF2Deriv, dF1Deriv(1, i), dF2Deriv(1, j))

                        ' calculating diagonal step
                        ' delta_uv - get cost factor for penalty of different step types (diagonal, right, down)
                        dblStpTypeCost(0) = dblMatrix(i - 1, j - 1) + dblStepCost

                        ' calculating down step
                        If intDownSteps(i - 1, j) < intConstr Then
                            dblStpTypeCost(1) = dblMatrix(i - 1, j) + dblStepCost '* 2
                        Else
                            dblStpTypeCost(1) = 10 ^ 300
                        End If

                        ' calculating right step
                        If intRightSteps(i, j - 1) < intConstr Then
                            dblStpTypeCost(2) = dblMatrix(i, j - 1) + dblStepCost '* 2
                        Else
                            dblStpTypeCost(2) = 10 ^ 300
                        End If

                        ' choose best type of step
                        ' Type of the Steps 0 - Down, 1 - Right, 2 - Diagonal
                        Dim intStepType As Integer = Calc_ArgMin(dblStpTypeCost, 3)

                        ' assign values
                        Select Case intStepType
                            Case DDOWN
                                ' constraint of the count of down steps
                                If (intDownSteps(i - 1, j) + 1) < 0.01 * lngNumOfIntervals Then
                                    dblMatrix(i, j) = dblStpTypeCost(intStepType)
                                    lngPath(i, j) = intStepType
                                    intDownSteps(i, j) = intDownSteps(i - 1, j) + 1
                                    intDiagSteps(i, j) = 0
                                    intRightSteps(i, j) = 0
                                    intPathLength(i, j) = intPathLength(i - 1, j) + 1
                                Else
                                    dblMatrix(i, j) = dblStpTypeCost(DDIAG)
                                    lngPath(i, j) = DDIAG
                                    intDownSteps(i, j) = 0
                                    intDiagSteps(i, j) = intDiagSteps(i - 1, j - 1) + 1
                                    intRightSteps(i, j) = 0
                                    intPathLength(i, j) = intPathLength(i - 1, j - 1) + 1
                                End If

                            Case DRIGHT
                                ' constraint of the count of right steps
                                If intRightSteps(i, j - 1) + 1 < 0.01 * lngNumOfIntervals Then
                                    dblMatrix(i, j) = dblStpTypeCost(intStepType)
                                    lngPath(i, j) = intStepType
                                    intDownSteps(i, j) = 0
                                    intDiagSteps(i, j) = 0
                                    intRightSteps(i, j) = intRightSteps(i, j - 1) + 1
                                    intPathLength(i, j) = intPathLength(i, j - 1) + 1
                                Else
                                    dblMatrix(i, j) = dblStpTypeCost(DDIAG)
                                    lngPath(i, j) = DDIAG
                                    intDownSteps(i, j) = 0
                                    intDiagSteps(i, j) = intDiagSteps(i - 1, j - 1) + 1
                                    intRightSteps(i, j) = 0
                                    intPathLength(i, j) = intPathLength(i - 1, j - 1) + 1
                                End If

                            Case DDIAG
                                dblMatrix(i, j) = dblStpTypeCost(intStepType)
                                lngPath(i, j) = intStepType
                                intDownSteps(i, j) = 0
                                intDiagSteps(i, j) = intDiagSteps(i - 1, j - 1) + 1
                                intRightSteps(i, j) = 0
                                intPathLength(i, j) = intPathLength(i - 1, j - 1) + 1
                        End Select
                    End If
                Next
            Next

            ' giving back the dtw paths for both arrays
            Calc_ComputePath = GetPath(lngPath, iDimF1, iDimF2, intPathLength(iDimF1 - 1, iDimF2 - 1),
                                        lngPath_1, lngPath_2, dblMatrix)
            PathLength = UBound(lngPath_1) + 1

        Catch ex As Exception
            iError = 1
            Dim sMsg As String = ErrMsg("Calc_ComputePath", "averdtw.dll",
                                         "Error at the evaluation of dynamic time warping path! " & ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Function

    ''' <summary>
    ''' Minimal value in one array
    ''' </summary>
    ''' <param name="dblValues">1-dim array</param>
    ''' <param name="intSize">size of the 1-dim array</param>
    ''' <returns>minimal value in the 1-dim array</returns>
    Private Function Calc_ArgMin(ByRef dblValues() As Double, ByVal intSize As Integer) As Integer

        ' index of minimal value
        Dim Index As Integer = 0
        Dim temp As Double = dblValues(0)

        For i As Integer = 1 To intSize - 1
            If dblValues(i) < temp Then
                temp = dblValues(i)
                Index = i
            End If
        Next i

        Return Index

    End Function

    ''' <summary>
    ''' Cost function from Wang and Gasser
    ''' </summary>
    ''' <param name="dblAlpha">parameter alfa from cost function</param>
    ''' <param name="dMaxUsedF1">max value of one component on curve 1</param>
    ''' <param name="dMaxUsedF2">max value of one component on curve 2</param>
    ''' <param name="dUsedF1">value of one component at one point on curve 1</param>
    ''' <param name="dUsedF2">value of one component at one point on curve 2</param>
    ''' <param name="dMaxUsedF1Deriv">max value der 1st derivative einer component of curve 1</param>
    ''' <param name="dMaxUsedF2Deriv">max value der 1st derivative einer component of curve 2</param>
    ''' <param name="dUsedF1Deriv">1st derivative of one component at one point on curve 1</param>
    ''' <param name="dUsedF2Deriv">1st derivative of one component at one point on curve 2</param>
    ''' <returns>value of the cost function by Wang and Gasser</returns>
    Private Function Calc_CostWG(ByVal dblAlpha As Double, ByVal dMaxUsedF1 As Double, ByRef dMaxUsedF2 As Double,
            ByRef dUsedF1 As Double, ByRef dUsedF2 As Double, ByVal dMaxUsedF1Deriv As Double, ByRef dMaxUsedF2Deriv As Double,
            ByRef dUsedF1Deriv As Double, ByRef dUsedF2Deriv As Double) As Double

        Try
            ' avoid the division through 0:
            If dMaxUsedF1 = 0 Then
                dMaxUsedF1 = 10 ^ -300
            End If
            If dMaxUsedF2 = 0 Then
                dMaxUsedF2 = 10 ^ -300
            End If
            If dMaxUsedF1Deriv = 0 Then
                dMaxUsedF1Deriv = 10 ^ -300
            End If
            If dMaxUsedF2Deriv = 0 Then
                dMaxUsedF2Deriv = 10 ^ -300
            End If

            '
            Dim dDist As Double = (dUsedF1 / dMaxUsedF1) - (dUsedF2 / dMaxUsedF2)
            dDist = dDist * dDist
            '
            Dim dDistDeriv As Double = (dUsedF1Deriv / dMaxUsedF1Deriv) - (dUsedF2Deriv / dMaxUsedF2Deriv)
            dDistDeriv = dDistDeriv * dDistDeriv

            Return (dblAlpha * dblAlpha * dDist + (1 - dblAlpha) * (1 - dblAlpha) * dDistDeriv)

        Catch ex As Exception
            Calc_CostWG = 0
            iError = 1
            Dim sMsg As String = ErrMsg("Calc_CostWG", "averdtw.dll", ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Function

    ''' <summary>
    '''  Evaluation of the Dynamic Time Warping (DTW) path
    '''      Input:    f1() And f2() - both curves for DTW-path
    '''               lngMatrix() - DTW matrix
    '''               f1Length And f2Length - length of the 1st And the 2nd arrays
    '''               intLength - length of the DTW path
    '''     Output:   lngPath_1() - DTW-path of function 1 (measurement number)
    '''               lngPath_2() - DTW-path of function  2 (measurement number)
    ''' </summary>
    ''' <param name="lngMatrix">DTW matrix</param>
    ''' <param name="f1Length">length of the 1st array</param>
    ''' <param name="f2Length">length of the 2nd array</param>
    ''' <param name="intLength">length of the DTW path</param>
    ''' <param name="lngPath_1">DTW-path of function 1 (measurement number)</param>
    ''' <param name="lngPath_2">DTW-path of function  2 (measurement number)</param>
    ''' <param name="dblMatrix"></param>
    ''' <returns></returns>
    Private Function GetPath(ByRef lngMatrix(,) As Long, ByRef f1Length As Integer,
                       ByRef f2Length As Integer, ByRef intLength As Integer, ByRef lngPath_1() As Long, ByRef lngPath_2() As Long,
                       ByRef dblMatrix(,) As Double) As Double

        Try

            ' dimension of help arrays
            Dim lngPath1(f1Length + f2Length) As Long
            Dim lngPath2(f1Length + f2Length) As Long

            Dim i As Integer = 0
            GetPath = dblMatrix(f1Length - 1, f2Length - 1)

            ' backtrack the path in the matrix
            While (f1Length >= 0 And f2Length >= 0)
                Select Case lngMatrix(f1Length, f2Length)
                    Case DDIAG
                        lngPath1(i) = f1Length
                        lngPath2(i) = f2Length

                        f1Length = f1Length - 1
                        f2Length = f2Length - 1

                    Case DDOWN
                        lngPath1(i) = f1Length
                        lngPath2(i) = f2Length

                        f1Length = f1Length - 1

                    Case DRIGHT
                        lngPath1(i) = f1Length
                        lngPath2(i) = f2Length

                        f2Length = f2Length - 1

                End Select

                intLength = intLength - 1
                i = i + 1
            End While

            ' save the paths in reversed order
            ReDim lngPath_1(0 To i - 1)
            ReDim lngPath_2(0 To i - 1)
            For j As Integer = 0 To i - 1
                ' path with measurement number
                lngPath_1(i - j - 1) = lngPath1(j)
                lngPath_2(i - j - 1) = lngPath2(j)
            Next j

        Catch ex As Exception
            iError = 1
            Dim sMsg As String = ErrMsg("GetPath", "averdtw.dll", ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Function

    ''' <summary>
    ''' Evaluation of the Dynamic Time Warping - paths of curves dblFunc1() and dblFunc2()
    ''' </summary>
    ''' <param name="dblAlpha"></param>
    ''' <param name="dblFunc1"></param>
    ''' <param name="dblFunc2"></param>
    ''' <param name="dblMax1"></param>
    ''' <param name="dblMax2"></param>
    ''' <param name="dF1Deriv"></param>
    ''' <param name="dF2Deriv"></param>
    ''' <param name="dblMaxDeriv1"></param>
    ''' <param name="dblMaxDeriv2"></param>
    ''' <param name="dblCurve1"></param>
    ''' <param name="dblCurve2"></param>
    ''' <param name="lngFile1"></param>
    ''' <param name="lngFile2"></param>
    ''' <param name="dblSmoothEqMed"></param>
    Private Sub Calc_AverTwoCurves(ByVal dblAlpha As Double, ByRef dblFunc1(,) As Double, ByRef dblFunc2(,) As Double,
                             ByRef dblMax1 As Double, ByRef dblMax2 As Double, ByRef dF1Deriv(,) As Double,
                             ByRef dF2Deriv(,) As Double, ByRef dblMaxDeriv1 As Double, ByRef dblMaxDeriv2 As Double,
                             ByRef dblCurve1(,) As Double, ByRef dblCurve2(,) As Double, ByVal lngFile1 As Long,
                             ByVal lngFile2 As Long, ByRef dblSmoothEqMed(,) As Double)

        Try

            Dim lngPath1() As Long              ' dtw path of curve 1 (measurement number)
            Dim lngPath2() As Long              ' dtw path of curve 2 (measurement number)

            ' calculation of dtw path
            ReDim lngPath1(0)
            ReDim lngPath2(0)
            Dim PathLength As Long = 0             ' lenght of dtw path
            ' cost function
            Dim dblCost As Double = Calc_ComputePath(dblAlpha, dblFunc1, dblFunc2, dblMax1, dblMax2, dF1Deriv, dF2Deriv,
                             dblMaxDeriv1, dblMaxDeriv2, lngPath1, lngPath2, PathLength)

            ' time step of average curve
            Dim dblTimeStep As Double = (dblCurve1(0, dblCurve1.GetUpperBound(1)) + dblCurve2(0, dblCurve2.GetUpperBound(1))) / 2

            ' dimension of the array for average curve
            Dim DTWCurveMed(dblCurve1.GetUpperBound(0), lngPath1.GetUpperBound(0)) As Double

            ' calculate the average curve
            ' go through the dtw path
            For j As Integer = 0 To lngPath1.GetUpperBound(0)
                ' normalized time: for many curves = ((n-1)/n * TimeCurve1 + 1/n * TimeCurve2)
                DTWCurveMed(0, j) = ((lngFile1 + 1) * lngPath1(j) / (lngFile2 + 1) + lngPath2(j) / (lngFile2 + 1))

                ' all components
                For i As Integer = 1 To UBound(dblCurve1, 1)
                    DTWCurveMed(i, j) = ((lngFile1 + 1) * dblCurve1(i, CInt(lngPath1(j))) /
                                         (lngFile2 + 1) + dblCurve2(i, CInt(lngPath2(j))) / (lngFile2 + 1))
                Next
            Next

            ' equidistant segmenting of average curve
            Calc_CurveEquidist(DTWCurveMed, dblSmoothEqMed)

            ' maximal time stemp
            Dim dblTimeMax1 As Double = dblFunc1(0, dblFunc1.GetUpperBound(1))
            Dim dblTimeMax2 As Double = dblFunc2(0, dblFunc2.GetUpperBound(1))

            ' average value of time for many curves = ((n-1)/n * f(t) + 1/n * g(t))
            ' calculating the time points
            dblTimeStep = ((lngFile1 + 1) * dblTimeMax1 / (lngFile2 + 1) +
                           dblTimeMax2 / (lngFile2 + 1)) / lngNumOfIntervals

            For j As Integer = 0 To lngNumOfIntervals
                dblSmoothEqMed(0, j) = dblSmoothEqMed(0, j) * dblTimeStep
            Next

        Catch ex As Exception
            iError = 1
            Dim sMsg As String = ErrMsg("Calc_AverTwoCurves", "averdtw.dll", ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Sub

    ''' <summary>
    ''' Calculate the dtw path between each original curve and the reference curve
    ''' </summary>
    Private Sub Calc_Dtw2AverCurve()

        Dim dblFunc1(,) As Double            ' 2-dim array with time and values of one component(curve 1)
        Dim dblFunc2(,) As Double            ' 2-dim array with time and values of one component(curve 2)
        Dim dF1Deriv(,) As Double            ' 2-dim array with time and values of 1st derivative for one component(curve 1)
        Dim dF2Deriv(,) As Double            ' 2-dim array with time and values of 1st derivative for one component(curve 2)
        Dim DTWCurveMed(,) As Double         ' temporary average curve

        ' dtw paths
        Dim lngPath1() As Long               ' dtw path for curve 1
        Dim lngPath2() As Long               ' dtw path for curve 1
        Dim lngPathAver(,) As Long           ' help array
        Dim lngPathOrig(,) As Long           ' help array

        Dim dblCurve1(,) As Double           ' 2-dim array with time and values of all component for curve 1
        Dim dblCurve2(,) As Double           ' 2-dim array with time and values of all component for curve 2
        Dim arrMax(,) As Double              ' array wit max values for all original curves
        Dim arrMin(,) As Double              ' array wit min values for all original curves
        Dim arrMedian(,) As Double           ' array wit median values for all original curves
        Dim arr25Perc(,) As Double           ' array wit 25 percentil values for all original curves
        Dim arr75Perc(,) As Double           ' array wit 75 percentil values for all original curves
        Dim CostDTW() As Double              ' cost function
        Dim PathLength() As Long             ' path lenght

        Try

            ReDim lngPathOrig(0, 0)
            ReDim lngPathAver(0, 0)

            ' 1. Step: DTW path of each original curve with the reference curve

            ' preparation: calculate of 1st derivative for one component of the refernce curve
            ReDim dblFunc1(0, 0)
            ReDim dF1Deriv(0, 0)
            Dim dblMax1 As Double = 0                ' max value of one component on the curve 1
            Dim dblMax2 As Double = 0               ' max value of one component on the curve 2
            Dim dblMaxDeriv1 As Double = 0           ' max value of 1st derivative for one component on the curve 1
            Dim dblMaxDeriv2 As Double = 0       ' max value of 1st derivative for one component on the curve 2
            Get2Dim_1CompForDTW(AverCurve, DerivAverCurve, AverMaxValue, AverMaxDeriv,
                                    dblFunc1, dF1Deriv, dblMax1, dblMaxDeriv1, intComponent)

            ReDim dblCurve1(0, 0)
            Copy2Dim_To_2Dim(AverCurve, dblCurve1)

            ' count of the original curves
            Dim lngFiles As Integer = UBound(OrigSmEq, 3)
            ReDim CostDTW(lngFiles)
            ReDim PathLength(lngFiles)

            ' all original curves
            For lngFile As Integer = 0 To lngFiles
                ' preparation: calculation of the 1st derivative for one component of one original curve
                ReDim dblFunc2(0, 0)
                ReDim dF2Deriv(0, 0)
                ReDim lngPath1(0)
                ReDim lngPath2(0)
                Get3Dim_1CompForDTW(OrigSmEq, dblDerivOrig, MaxValue, MaxDeriv,
                                    dblFunc2, dF2Deriv, dblMax2, dblMaxDeriv2, CInt(lngFile), intComponent)

                Dim dblWGAlpha As Double             ' parameter
                Calc_Alfa(dblWGAlpha, dblFunc1, dblFunc2, dblMax1, dblMax2, dF1Deriv, dF2Deriv, dblMaxDeriv1,
                                     dblMaxDeriv2)

                ' calculate the dtw path
                CostDTW(lngFile) = Calc_ComputePath(dblWGAlpha, dblFunc1, dblFunc2, dblMax1, dblMax2, dF1Deriv, dF2Deriv,
                             dblMaxDeriv1, dblMaxDeriv2, lngPath1, lngPath2, PathLength(lngFile))

                ReDim dblCurve2(0, 0)
                Get3Dim_AllComp(OrigSmEq, dblCurve2, CInt(lngFile))

                ' save the dtw path
                If lngFile = 0 Then
                    'first loop, empty arrays
                    ReDim lngPathAver(lngFiles, lngPath1.GetUpperBound(0))
                    ReDim lngPathOrig(lngFiles, lngPath1.GetUpperBound(0))
                Else
                    If lngPath1.GetUpperBound(0) > lngPathOrig.GetUpperBound(1) Then
                        ReDim Preserve lngPathAver(lngFiles, lngPath1.GetUpperBound(0))
                        ReDim Preserve lngPathOrig(lngFiles, lngPath1.GetUpperBound(0))
                    End If
                End If

                For lngStep As Integer = 0 To UBound(lngPath1)
                    lngPathAver(lngFile, lngStep) = lngPath1(lngStep)
                    lngPathOrig(lngFile, lngStep) = lngPath2(lngStep)
                Next lngStep

                Application.DoEvents()

            Next

            ReDim DTWCurveMed(0, 0)
            ReDim arrMax(0, 0)
            ReDim arrMin(0, 0)
            ReDim arrMedian(0, 0)
            ReDim arr25Perc(0, 0)
            ReDim arr75Perc(0, 0)
            Calc_AverFromPaths(lngPathAver, lngPathOrig, OrigSmEq, DTWCurveMed,
                                arrMax, arrMin, arrMedian, arr25Perc, arr75Perc, CInt(lngFiles))

            ' evaluation of the equdistant average curve with 1st derivative
            Calc_AverCurveEq(DTWCurveMed, CInt(lngNumOfIntervals))

            ' max/min, ... curves
            Calc_CurveEquidist(arrMax, MaxCurve)
            Calc_CurveEquidist(arrMin, MinCurve)
            Calc_CurveEquidist(arrMedian, MedianCurve)
            Calc_CurveEquidist(arr25Perc, Perc25Curve)
            Calc_CurveEquidist(arr75Perc, Perc75Curve)

            ' smooth the average curve
            Calc_AverSmooth(dblSmooth)
            Calc_QuartileMaxMin_Smooth()

        Catch ex As Exception
            iError = 1
            Dim sMsg As String = ErrMsg("Calc_Dtw2AverCurve", "averdtw.dll",
                                        "Error in the calculation of the average curve! " & ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Sub

    ''' <summary>
    ''' Saving a 2dim array dblCurveAlt() in the other 2-dim array dblCurveNeu()
    ''' </summary>
    ''' <param name="dblCurveAlt"></param>
    ''' <param name="dblCurveNeu"></param>
    Private Sub Copy2Dim_To_2Dim(ByRef dblCurveAlt(,) As Double, ByRef dblCurveNeu(,) As Double)

        Try
            ' redimension the array
            ReDim dblCurveNeu(0 To UBound(dblCurveAlt, 1), 0 To UBound(dblCurveAlt, 2))

            ' save the values
            For i As Integer = 0 To UBound(dblCurveAlt, 2)
                For j As Integer = 0 To UBound(dblCurveAlt, 1)
                    dblCurveNeu(j, i) = dblCurveAlt(j, i)
                Next j
            Next i

        Catch ex As Exception
            iError = 1
            Dim sMsg As String = ErrMsg("Copy2Dim_To_2Dim", "averdtw.dll",
                                        "Error at saving of smoothed (temporary) averaged data!! " & ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Sub

    ''' <summary>
    ''' Calculate the average curve (see paper)
    ''' </summary>
    ''' <param name="lngPathAver"></param>
    ''' <param name="lngPathOrig"></param>
    ''' <param name="dblOrigN"></param>
    ''' <param name="CurveMed"></param>
    ''' <param name="arrMax"></param>
    ''' <param name="arrMin"></param>
    ''' <param name="arrMedian"></param>
    ''' <param name="arr25Perc"></param>
    ''' <param name="arr75Perc"></param>
    ''' <param name="lngFiles"></param>
    Private Sub Calc_AverFromPaths(ByRef lngPathAver(,) As Long, ByRef lngPathOrig(,) As Long,
                                  ByRef dblOrigN(,,) As Double, ByRef CurveMed(,) As Double,
                                  ByRef arrMax(,) As Double, ByRef arrMin(,) As Double,
                                  ByRef arrMedian(,) As Double, ByRef arr25Perc(,) As Double,
                                  ByRef arr75Perc(,) As Double, ByRef lngFiles As Long)

        Try

            ' points of the reference curve
            Dim lngAnzReferenz As Integer = AverCurve.GetUpperBound(1)
            ' dimension the array of average curve + max/min/median/perc25/perc75
            ReDim CurveMed(dblOrigN.GetUpperBound(0), lngAnzReferenz)
            ReDim arrMax(dblOrigN.GetUpperBound(0), lngAnzReferenz)
            ReDim arrMin(dblOrigN.GetUpperBound(0), lngAnzReferenz)
            ReDim arrMedian(dblOrigN.GetUpperBound(0), lngAnzReferenz)
            ReDim arr25Perc(dblOrigN.GetUpperBound(0), lngAnzReferenz)
            ReDim arr75Perc(dblOrigN.GetUpperBound(0), lngAnzReferenz)

            ' count of points of average curve
            Dim lngMedPoint As Integer = 0
            ' the number of current points of the average curve
            Dim lngAnzMed As Integer = lngAnzReferenz

            ' count of the original files in the evaluation
            Dim lngAnzFiles As Integer = UBound(OrigSmEq, 3)
            ' lenght of array lngPathAver()
            Dim lngLaengepath As Integer = UBound(lngPathAver, 2)

            Dim i As Integer = 0
            Dim lngSteps As Integer = 0        ' max count of original curve points connected to one point of reference curve
            Dim arrPunktAkt(,) As Integer   ' path segment of the original curves
            Dim arrPunktNext() As Integer   ' the next path segment of the original curves

            ' reference curve: point of point
            For lngPoint As Integer = 0 To lngAnzReferenz
                ' redim arrays
                ReDim arrPunktAkt(lngAnzFiles, 0)
                ReDim arrPunktNext(lngAnzFiles)
                lngSteps = UBound(arrPunktAkt, 2)

                ' points on the original curves, connected with
                ' the current point on the reference curve
                For lngFile As Integer = 0 To lngAnzFiles
                    For lngAverpath As Integer = 0 To lngLaengepath
                        If lngPathAver(lngFile, lngAverpath) = lngPoint Then
                            i = -1
                            Do While lngPathAver(lngFile, lngAverpath) = lngPoint
                                i = i + 1
                                If i > lngSteps Then
                                    lngSteps = arrPunktAkt.GetUpperBound(1) + 1
                                    ReDim Preserve arrPunktAkt(lngAnzFiles, lngSteps)
                                End If

                                arrPunktAkt(lngFile, i) = CInt(lngPathOrig(lngFile, lngAverpath))
                                lngAverpath = lngAverpath + 1
                                If lngAverpath > lngLaengepath Then Exit Do
                            Loop

                            ' the next point on the reference curve
                            If lngAverpath < lngLaengepath + 1 Then
                                If lngPathOrig(lngFile, lngAverpath) <> 0 Then
                                    arrPunktNext(lngFile) = CInt(lngPathOrig(lngFile, lngAverpath))
                                Else
                                    arrPunktNext(lngFile) = CInt(lngPathOrig(lngFile, lngAverpath - 1))
                                End If
                            Else
                                arrPunktNext(lngFile) = CInt(lngPathOrig(lngFile, lngAverpath - 1))
                            End If

                            Exit For
                        End If
                    Next lngAverpath
                    ' the next original curve
                Next lngFile

                ' calculate a segment of the average curve
                ' redim arrays
                If lngMedPoint + lngSteps > lngAnzMed Then
                    lngAnzMed = lngMedPoint + lngSteps
                    ReDim Preserve CurveMed(dblOrigN.GetUpperBound(0), lngAnzMed)
                    ReDim Preserve arrMax(dblOrigN.GetUpperBound(0), lngAnzMed)
                    ReDim Preserve arrMin(dblOrigN.GetUpperBound(0), lngAnzMed)
                    ReDim Preserve arrMedian(dblOrigN.GetUpperBound(0), lngAnzMed)
                    ReDim Preserve arr25Perc(dblOrigN.GetUpperBound(0), lngAnzMed)
                    ReDim Preserve arr75Perc(dblOrigN.GetUpperBound(0), lngAnzMed)
                End If

                For j As Integer = lngMedPoint To lngMedPoint + lngSteps
                    For intComp As Integer = 0 To UBound(dblOrigN, 1)
                        Dim dblSumme As Double = 0
                        For lngFile = 0 To lngAnzFiles
                            dblSumme = dblSumme + dblOrigN(intComp, arrPunktAkt(lngFile, 0), lngFile) +
                                       (j - lngMedPoint) * (dblOrigN(intComp, arrPunktNext(lngFile), lngFile) -
                                       dblOrigN(intComp, arrPunktAkt(lngFile, 0), lngFile)) / (lngSteps + 1)
                        Next lngFile
                        CurveMed(intComp, j) = dblSumme / (lngFiles + 1)

                        If intComp > 0 Then
                            ' max/min/median/25perc/75perc
                            Dim dblHilf(0 To lngAnzFiles) As Double
                            For lngFile = 0 To lngAnzFiles
                                dblHilf(lngFile) = dblOrigN(intComp, arrPunktAkt(lngFile, 0), lngFile) +
                                           (j - lngMedPoint) * (dblOrigN(intComp, arrPunktNext(lngFile), lngFile) -
                                           dblOrigN(intComp, arrPunktAkt(lngFile, 0), lngFile)) / (lngSteps + 1)
                            Next lngFile

                            Dim dblSpannW As Double = 0
                            Dim lngMaxNr As Integer = 0
                            Dim lngMinNr As Integer = 0
                            Calc_MaxMin(dblHilf, arrMax(intComp, j), arrMin(intComp, j), dblSpannW, CLng(lngMaxNr), CLng(lngMinNr))
                            If iError = 0 Then arrMedian(intComp, j) = Calc_Median(dblHilf)

                            Dim arrQuartil(0 To 3) As Double
                            If Not (Calc_Quartil(dblHilf, arrQuartil)) Then Exit Sub

                            arr25Perc(intComp, j) = arrQuartil(1)
                            arr75Perc(intComp, j) = arrQuartil(3)
                        Else
                            arrMax(intComp, j) = CurveMed(intComp, j)
                            arrMin(intComp, j) = CurveMed(intComp, j)
                            arrMedian(intComp, j) = CurveMed(intComp, j)
                            arr25Perc(intComp, j) = CurveMed(intComp, j)
                            arr75Perc(intComp, j) = CurveMed(intComp, j)
                        End If

                    Next
                Next

                lngMedPoint = lngMedPoint + lngSteps + 1
            Next

        Catch ex As Exception
            iError = 1
            Dim sMsg As String = ErrMsg("Calc_AverFromPaths", "averdtw.dll",
                                        "Error at the evaluation of dynamic time warping path! " & ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Sub

    ''' <summary>
    ''' Equidist time segmenting the average curve
    ''' </summary>
    ''' <param name="dblCurveAver"></param>
    ''' <param name="lngNumEquidist">count of the equidistant intervals</param>
    Private Sub Calc_AverCurveEq(ByRef dblCurveAver(,) As Double, ByRef lngNumEquidist As Long)

        Try
            Dim dblCurveMed(,) As Double     ' array with averaged smoothed data

            ' Equidist time segmenting of average curve
            ReDim dblCurveMed(0, 0)
            Calc_CurveEquidist(dblCurveAver, dblCurveMed)

            ' redimension of all average arrays
            ReDim AverCurve(dblCurveMed.GetUpperBound(0), CInt(lngNumEquidist))                ' average curve
            ReDim DerivAverCurve(dblCurveMed.GetUpperBound(0), CInt(lngNumEquidist))         ' 1st derivativeen of average curve
            ReDim AverMaxValue(dblCurveMed.GetUpperBound(0))
            ReDim AverMaxDeriv(dblCurveMed.GetUpperBound(0))

            ' save
            For intComp As Integer = 0 To dblCurveMed.GetUpperBound(0)
                For intPoint As Integer = 0 To dblCurveMed.GetUpperBound(1)
                    AverCurve(intComp, intPoint) = dblCurveMed(intComp, intPoint)
                Next
            Next

            ' calculate the 1st derivative
            Calc_DerivNum(AverCurve, DerivAverCurve)
            ' max values and max value of 1st derivative
            Search_Max(AverCurve, AverMaxValue)
            Search_Max(DerivAverCurve, AverMaxDeriv)

        Catch ex As Exception
            iError = 1
            Dim sMsg As String = ErrMsg("Calc_AverCurveEq", "averdtw.dll", ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Sub

    ''' <summary>
    ''' Smoothing of the average curve
    ''' </summary>
    ''' <param name="dblSmoothing"></param>
    Private Sub Calc_AverSmooth(ByRef dblSmoothing As Double)

        Try

            ' redimension of all arrays for average curve
            ReDim AverCurveSm(AverCurve.GetUpperBound(0), lngNumOfIntervals)
            ReDim DerivAverCurveSm(DerivAverCurve.GetUpperBound(0), lngNumOfIntervals)
            ReDim AverMaxValueSm(AverMaxValue.GetUpperBound(0))
            ReDim AverMaxDerivSm(AverMaxDeriv.GetUpperBound(0))

            ' smoothing of the average curve
            Calc_CurveSmoothing(AverCurve, AverCurveSm, dblSmooth)
            Calc_DerivNum(AverCurveSm, DerivAverCurveSm)
            ' max values of each component and its 1st derivative
            Search_Max(AverCurveSm, AverMaxValueSm)
            Search_Max(DerivAverCurveSm, AverMaxDerivSm)

        Catch ex As Exception
            iError = 1
            Dim sMsg As String = ErrMsg("Calc_AverSmooth", "averdtw.dll", ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Sub

    ''' <summary>
    ''' Smoothing of min/max/median/25 percentil/75 percentil - curves
    ''' </summary>
    Private Sub Calc_QuartileMaxMin_Smooth()

        Calc_CurveSmoothing(MaxCurve, MaxCurveSm, dblSmoothMax)
        Calc_CurveSmoothing(MinCurve, MinCurveSm, dblSmoothMin)
        Calc_CurveSmoothing(MedianCurve, MedianCurveSm, dblSmoothMedian)
        Calc_CurveSmoothing(Perc25Curve, Perc25CurveSm, dblSmooth25Perc)
        Calc_CurveSmoothing(Perc75Curve, Perc75CurveSm, dblSmooth75Perc)

    End Sub

    ''' <summary>
    ''' Calculate the min and max value in an array
    ''' Input: arrWerte() - 1-dim array
    ''' Output: min value, index of min value in the array,
    ''' max value,  index of max value in the array, range
    ''' </summary>
    ''' <param name="arrWerte">1-dim array</param>
    ''' <param name="dblMaxVal">max value</param>
    ''' <param name="dblMinVal">min value</param>
    ''' <param name="dblSpannweite">range value</param>
    ''' <param name="lngMaxNr"> index of max value in the array</param>
    ''' <param name="lngMinNr"> index of min value in the array</param>
    Private Sub Calc_MaxMin(ByRef arrWerte() As Double, ByRef dblMaxVal As Double,
                              ByRef dblMinVal As Double, ByRef dblSpannweite As Double,
                              ByRef lngMaxNr As Long, ByRef lngMinNr As Long)

        Try
            ' control
            If UBound(arrWerte) = 0 Then
                iError = 1
                dblMinVal = 0
                lngMinNr = 0
                dblMaxVal = 0
                lngMaxNr = 0
                dblSpannweite = 0
                Exit Sub       '
            End If

            ' evaluate upper and lower bound
            Dim intUNumber As Integer = UBound(arrWerte, 1)
            Dim intLNumber As Integer = LBound(arrWerte, 1)

            ' initialize
            dblMinVal = 10 ^ 300
            dblMaxVal = -10 ^ 300

            For i As Integer = intLNumber To intUNumber
                ' min value + index
                If dblMinVal > arrWerte(i) Then
                    dblMinVal = arrWerte(i)
                    lngMinNr = i
                End If

                ' max value with index
                If dblMaxVal < arrWerte(i) Then
                    dblMaxVal = arrWerte(i)
                    lngMaxNr = i
                End If
            Next i

            dblSpannweite = dblMaxVal - dblMinVal

        Catch ex As Exception
            dblMinVal = 0
            lngMinNr = 0
            dblMaxVal = 0
            lngMaxNr = 0
            dblSpannweite = 0
            iError = 1
            Dim sMsg As String = ErrMsg("Calc_MaxMin", "averdtw.dll", ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Sub

    ''' <summary>
    ''' Calculate of the median value of a array
    ''' Evaluation: 50% - the number of the  median for the curve: 1 + (n-1) * 0,50, n - count of curves
    ''' </summary>
    ''' <param name="arrWerte">1-dim array</param>
    ''' <returns>median value of arrWerte()</returns>
    Private Function Calc_Median(ByRef arrWerte() As Double) As Double

        Dim arrHilf() As Double     ' help array
        Dim dbl50Perc As Double     ' median value
        Dim int50Perc As Integer    ' order number of median
        Dim dblRest As Double

        Try

            ' control
            If UBound(arrWerte) = 0 Then
                iError = 1
                Calc_Median = 0
                Throw New Exception("Error in Calc_Median: Array is empty")
            End If

            ' count of the elements and upper and lower bound
            Dim intUNumber As Integer = UBound(arrWerte, 1)
            Dim intLNumber As Integer = LBound(arrWerte, 1)
            Dim intNumber As Integer = intUNumber - intLNumber + 1

            If intNumber = 1 Then
                ' for one file
                Calc_Median = arrWerte(intUNumber)
            Else
                ' dimension of help array
                ReDim arrHilf(0 To intUNumber)
                ' copy values in a help array
                For i As Integer = intLNumber To intUNumber
                    arrHilf(i) = arrWerte(i)
                Next i

                ' sort the help array
                If Not (Calc_QuickSort(arrHilf, intLNumber, intUNumber)) Then
                    Throw New Exception("Error in Calc_QuickSort")
                End If

                ' 50 percentil value
                dbl50Perc = CDbl(1 + (intNumber - 1) * 0.5)

                ' median
                dblRest = dbl50Perc - CLng(dbl50Perc) \ 1
                If dblRest < 0 Then dblRest = 1 + dblRest
                int50Perc = CInt(dbl50Perc - dblRest)
                Calc_Median = arrHilf(int50Perc - 1) + dblRest * (arrHilf(int50Perc) - arrHilf(int50Perc - 1))

            End If

        Catch ex As Exception
            Calc_Median = 0
            iError = 1
            Dim sMsg As String = ErrMsg("Calc_Median", "averdtw.dll", ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Function

    ''' <summary>
    ''' Calculation of the quartil values of an array
    '''     Evaluation form:
    '''     0% - the number of the 0-pesentil for the curve: 1 + (n-1) * 0,00, n - count of curves
    '''     25% - the number of the  25-pesentil for the curve: 1 + (n-1) * 0,25, n - count of curves
    '''     75% - the number of the  75-pesentil for the curve: 1 + (n-1) * 0,75, n - count of curves
    '''     100% - the number of the  100-pesentil for the curve: 1 + (n-1) * 1,00, n - count of curves
    ''' </summary>
    ''' <param name="arrWerte">1-dim array</param>
    ''' <param name="arrQuartil">output array with 25-, 50-, and 75-percentils</param>
    Private Function Calc_Quartil(ByRef arrWerte() As Double, ByRef arrQuartil() As Double) As Boolean

        Try

            ' control
            If UBound(arrWerte) = 0 Then
                Throw New Exception("Error in Calc_Quartil(): array is empty")       ' no values
            End If

            ' count der elements and upper and lower bound
            Dim intUNumber As Integer = UBound(arrWerte, 1)
            Dim intLNumber As Integer = LBound(arrWerte, 1)
            Dim intNumber As Integer = intUNumber - intLNumber + 1

            ReDim arrQuartil(0 To 3)
            If intNumber = 1 Then
                ' for one file
                arrQuartil(1) = arrWerte(intUNumber)
                arrQuartil(2) = 0
                arrQuartil(3) = arrWerte(intUNumber)
            Else
                ' dimention of help array
                Dim arrHilf(0 To intUNumber) As Double

                ' copy values in a help array
                For i As Integer = intLNumber To intUNumber
                    arrHilf(i) = arrWerte(i)
                Next

                ' sort the help array
                If Not (Calc_QuickSort(arrHilf, intLNumber, intUNumber)) Then
                    Throw New Exception("Error in Calc_QuickSort")
                End If

                ' 25 and 75 percentil value
                Dim dbl25Perc As Double = CDbl(1 + (intNumber - 1) * 0.25)
                Dim dbl75Perc As Double = CDbl(1 + (intNumber - 1) * 0.75)

                ' 25 percentil
                Dim dblRest As Double = dbl25Perc - CLng(dbl25Perc) \ 1
                If dblRest < 0 Then dblRest = 1 + dblRest
                Dim int25Perc As Integer = CInt(dbl25Perc - dblRest)
                arrQuartil(1) = arrHilf(int25Perc - 1) + dblRest * (arrHilf(int25Perc) - arrHilf(int25Perc - 1))

                ' 75 percentil
                dblRest = dbl75Perc - CLng(dbl75Perc) \ 1
                If dblRest < 0 Then dblRest = 1 + dblRest
                Dim int75Perc As Integer = CInt(dbl75Perc - dblRest)
                arrQuartil(3) = arrHilf(int75Perc - 1) + dblRest * (arrHilf(int75Perc) - arrHilf(int75Perc - 1))

                ' half quartil
                arrQuartil(2) = arrQuartil(3) - arrQuartil(1)
            End If
            Return True

        Catch ex As Exception
            For i = 1 To 3 : arrQuartil(i) = 0 : Next i
            iError = 1
            Dim sMsg As String = ErrMsg("Calc_Quartil", "averdtw.dll", ex.Message)
            Throw New Exception(sMsg, ex)
        End Try
    End Function

    ''' <summary>
    ''' QuickSort algorithm
    ''' vSort() : array that has To be sorted
    ''' lngStart, lngEnd: part of the array to sort
    ''' </summary>
    ''' <param name="vSort">array that has To be sorted</param>
    ''' <param name="lngStart">start of the array to sort</param>
    ''' <param name="lngEnd">end of the array to sort</param>
    Private Function Calc_QuickSort(ByRef vSort() As Double, Optional ByVal lngStart As Integer = -1,
                     Optional ByVal lngEnd As Integer = -1) As Boolean

        Try

            If lngStart = -1 Then lngStart = vSort.GetLowerBound(0)
            If lngEnd = -1 Then lngEnd = vSort.GetUpperBound(0)
            Dim i As Integer = lngStart
            Dim j As Integer = lngEnd
            Dim X As Double = vSort((lngStart + lngEnd) \ 2)

            ' segmenting the array
            Do

                While (vSort(i) < X) : i = i + 1 : End While
                While (vSort(j) > X) : j = j - 1 : End While

                If (i <= j) Then
                    ' interchange the values
                    Dim h As Double = vSort(i)
                    vSort(i) = vSort(j)
                    vSort(j) = h
                    i = i + 1 : j = j - 1
                End If
            Loop Until (i > j)

            ' recursiv process
            If (lngStart < j) Then Calc_QuickSort(vSort, lngStart, j)
            If (i < lngEnd) Then Calc_QuickSort(vSort, i, lngEnd)

            Return True

        Catch ex As Exception
            Return False
        End Try

    End Function

    ''' <summary>
    ''' String for the runtime error message
    ''' </summary>
    ''' <param name="strProzedur">dll function/sub</param>
    ''' <param name="strModul">dll modul</param>
    ''' <param name="strDescr">runtime error message</param>
    Private Function ErrMsg(ByVal strProzedur As String, ByVal strModul As String, ByVal strDescr As String) As String
        Return ("Error in the function " & strProzedur & " in " & strModul & ": " & vbCrLf &
               strDescr & ", Runtime-Error")
    End Function
End Class
